from types import ModuleType
from typing import Dict, List
import pytest

class Item:
    def __init__(self, item: pytest.Item) -> None:
        self._item = item

    @property
    def description(self):        
        description = self._item.obj.__doc__

        if not description:
            description = self._item.name
            description = self._parse_description(description)

        description = description.splitlines()[0]
        description = description.capitalize()
        return description

    def _parse_description(self, description: str):
        split = False
        if description.lower().startswith("it_"):
            description = description.replace('it_', '')
            split = True

        if description.lower().startswith("test_it_"):
            description = description.replace('test_it_', '')
            split = True

        if description.lower().startswith("test_describe_"):
            description = description.replace('test_describe_', '')
            split = True

        if description.lower().startswith("describe_"):
            description = description.replace('describe_', '')
            split = True

        if split:
            description = description.replace('_', ' ')

        return description

    def __repr__(self):
        return self._item.__repr__()


class Test(Item):
    def __init__(self, item: pytest.Item) -> None:
        super().__init__(item)
        self.container: Container = None
        self.outcome: str = None

    @property
    def level(self) -> int:
        return self.container.level +1
   
    
class Container(Item):
    def __init__(self, item: pytest.Item):
        super().__init__(item)
        self.tests: List[Test] = list()
        self.containers: List[Container] = list()
        self.parent = None

    def add(self, test: Test):
        self.tests.append(test)
        test.container = self

    def add_container(self, container: 'Container'):
        self.containers.append(container)
        container.parent = self

    def flat_list(self):
        containers = []
        container = self
        while container:
            containers.insert(0, container)
            container = container.parent
        return containers

    @property
    def level(self) -> int:
        level = 0
        
        item = self._item
        while item.parent and not isinstance(item.parent.obj, ModuleType):
            level += 1
            item = item.parent
        
        return level


class ItemFactory:
    def __init__(self) -> None:
        self.container_factory = ContainerFactory()

    def create(self, item: pytest.Item) -> Test:
        test_item = Test(item)

        container_item = item.parent
        container = self.container_factory.create(container_item)
        container.add(test_item)       

        return test_item
        

class ContainerFactory:
    def __init__(self) -> None:
        self.containers: Dict[str, Container] = dict()

    def create(self, item) -> Container:
        containers = self._create_containers(item)
        return containers[-1]
    
    def _create_unique_container(self, item):
        if item not in self.containers:
            container = Container(item)
            self.containers[item] = container

        container = self.containers.get(item)
        return container

    def _create_containers(self, item):
        containers = []
        child_container = None
        while item and not isinstance(item.obj, ModuleType):
            container = self._create_unique_container(item)
            if child_container:
                container.add_container(child_container)

            containers.insert(0, container)
            child_container = container
            item = item.parent
        return containers


import pytest
from .item import Container, Test


def print_container(container: Container):
    output = "\n"
    for container in container.flat_list():
        output += print_parent_container(container)
        container = container.parent
    return output

def print_parent_container(container: Container):
    ident = "  " * container.level
    output = f"\n{ident}{container.description}"
    return output

def print_test(test: Test):
    ident = "  " * test.level
    status = print_test_status(test)
    output = f"\n{ident}{status} {test.description}"
    return output

def print_test_status(test: Test):
    if test.outcome == 'passed':
        return '✓'
    elif test.outcome == 'failed':
        return '✗'
    else:
        return '»'



import pytest
from pytest_pyspec.item import ItemFactory, Test
from pytest_pyspec.output import print_container, print_test


def pytest_addoption(parser: pytest.Parser,
                     pluginmanager: pytest.PytestPluginManager):
    group = parser.getgroup('general')
    group.addoption(
        '--pyspec',
        action='store_true',
        dest='pyspec',
        help='Enables pyspec features'
    )


enabled = False
def pytest_configure(config: pytest.Config):
    global enabled
    if config.getoption('pyspec') and not config.getoption('verbose'):
        enabled = True

        python_functions = config.getini("python_functions")
        python_functions.append('it_')
        config.option.python_functions = python_functions

        python_classes = config.getini("python_classes")
        python_classes.append('describe_')
        python_classes.append('with_')
        config.option.python_classes = python_classes



test_key = pytest.StashKey[Test]()
prev_test_key = pytest.StashKey[Test]()
def pytest_collection_modifyitems(session, config, items):
    if enabled:
        factory = ItemFactory()
        prev_test = None
        for i, item in enumerate(items):
            test = factory.create(item)
            item.stash[test_key] = test
            item.stash[prev_test_key] = prev_test
            prev_test = test


@pytest.hookimpl(hookwrapper=True)
def pytest_runtest_makereport(item, call):
    outcome = yield
    if enabled:
        report: pytest.Report = outcome.get_result()
        #TODO Check whether the report has a stash
        #TODO move previous test to test class
        report.test = item.stash[test_key]
        report.prev_test = item.stash[prev_test_key]


def pytest_report_teststatus(report: pytest.TestReport, config: pytest.Config):
    if enabled:
        test = report.test
        prev_test = report.prev_test

        if report.when == 'setup':
            if not prev_test \
                    or test.container != prev_test.container:
                # Show container
                output = print_container(test.container)
                return '', output, ('', {'white': True})

        if report.when == 'call':
            test.outcome = report.outcome
            output = print_test(test)
            return report.outcome, output, ''
